import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/settings/clean_android_design_style.dart';
import 'package:clean_android_design/src/utils/item_section_widget.dart';
import 'package:flutter/material.dart';

import 'default_field_widget.dart';

class TextFieldWidget extends StatefulWidget implements IListItem {
  final bool isFirst;
  final bool isLast;
  final Text title;

  final void Function(String) saveValue;
  final String Function() getValue;

  final FieldValidation validation;
  final FocusNode focusNode;
  final FocusNode nextFocusNode;
  final void Function(String) onSubmit;
  final TextInputType keyboardType;
  final bool multiLine;
  final bool separationLine;
  final int maxLines;
  final bool isEditable;

  const TextFieldWidget({
    Key key,
    this.saveValue,
    this.title,
    @required this.getValue,
    this.validation,
    this.focusNode,
    this.nextFocusNode,
    this.onSubmit,
    this.keyboardType,
    this.multiLine = false,
    this.separationLine = true,
    this.maxLines = 8,
    @required this.isEditable,
  })  : assert(getValue != null),
        assert(!(isEditable ?? true) || saveValue != null),
        isFirst = false,
        isLast = false,
        super(key: key);

  const TextFieldWidget._copy({
    Key key,
    this.isFirst,
    this.isLast,
    this.title,
    this.saveValue,
    this.getValue,
    this.validation,
    this.isEditable,
    this.focusNode,
    this.nextFocusNode,
    this.onSubmit,
    this.keyboardType,
    this.multiLine = false,
    this.separationLine = true,
    this.maxLines,
  }) : super(key: key);

  @override
  _TextFieldWidgetState createState() => _TextFieldWidgetState();

  @override
  IListItem copyWithPosition({bool isFirst, bool isLast}) {
    return TextFieldWidget._copy(
      key: this.key,
      saveValue: this.saveValue,
      getValue: this.getValue,
      title: this.title,
      validation: this.validation,
      isEditable: this.isEditable,
      focusNode: this.focusNode,
      nextFocusNode: this.nextFocusNode,
      onSubmit: this.onSubmit,
      keyboardType: this.keyboardType,
      multiLine: this.multiLine,
      separationLine: this.separationLine,
      maxLines: this.maxLines,
      isFirst: isFirst ?? this.isFirst,
      isLast: isLast ?? this.isLast,
    );
  }
}

class _TextFieldWidgetState extends State<TextFieldWidget> {
  String errorMessage;
  FocusNode validationFocusNode;

  @override
  void initState() {
    super.initState();

    widget.validation?.connectToField(
        onValidate: (errorMessage) {
          if (mounted) setState(() => this.errorMessage = errorMessage);
        },
        focusNode: widget.focusNode ?? (validationFocusNode = FocusNode()));
  }

  @override
  Widget build(BuildContext context) {
    var text = widget.getValue?.call() ?? "";
    var controller = TextEditingController()
      ..value = TextEditingValue(
          text: text,
          selection: TextSelection(
              baseOffset: text.length, extentOffset: text.length));
    controller.addListener(() => widget.saveValue
        ?.call(controller.text.trim() == "" ? null : controller.text.trim()));

    var inputBorder = !widget.isLast && widget.separationLine
        ? UnderlineInputBorder(
            borderSide: BorderSide(
                color: CleanAndroidDesignDefaultColors.defaultColorLight,
                width: 1))
        : InputBorder.none;
    var focusedBorder = UnderlineInputBorder(
        borderSide: BorderSide(
      color: CleanAndroidDesignDefaultColors.accent,
      width: 2,
    ));

    return DefaultFieldWidget(
      isFirst: widget.isFirst,
      isLast: widget.isLast,
      separationLine: false,
      title: widget.title,
      childPadding:
          const EdgeInsets.only(left: 30, right: 30, bottom: 8, top: 3),
      child: TextField(
        style: TextStyles.textStyleSize(17)
            .copyWith(color: CleanAndroidDesignDefaultColors.black),
        decoration: InputDecoration(
          contentPadding: EdgeInsets.only(top: 2, bottom: 10),
          isDense: true,
          enabledBorder: inputBorder,
          disabledBorder: inputBorder,
          focusedBorder: focusedBorder,
          errorText: errorMessage,
        ),
        controller: controller,
        cursorColor: CleanAndroidDesignDefaultColors.accentLight,
        enabled: widget.isEditable,
        keyboardType: widget.keyboardType,
        maxLines: widget.multiLine ? widget.maxLines : 1,
        minLines: widget.multiLine
            ? min(max(1, controller.text.split("\n").length),
                (widget.maxLines ?? 100))
            : 1,
        textInputAction: widget.multiLine
            ? TextInputAction.newline
            : widget.nextFocusNode == null
                ? TextInputAction.done
                : TextInputAction.next,
        focusNode: widget.focusNode ?? validationFocusNode,
        onSubmitted: (value) {
          if (widget.nextFocusNode != null)
            FocusScope.of(context).requestFocus(widget.nextFocusNode);
          widget.onSubmit?.call(value);
        },
      ),
    );
  }
}
