import 'dart:async';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';

class ListScaffoldWidget<T extends IdItemInterface> extends StatefulWidget {
  final Widget title;
  final List<T> list;
  final List<Widget> appBarButtons;
  final bool outsourceMenuButtons;
  final ISelectableListItem Function(BuildContext context, State state, T item)
      itemBuilder;
  final Comparable Function(T) sortProperty;
  final List<Comparable> Function(T) sortProperties;
  final bool reverseSort;
  final Comparable Function(T) groupProperty;
  final Widget Function(T) groupTitle;
  final List<Widget> Function(BuildContext context,
          ListScaffoldWidgetState<T> listScaffoldWidgetState)
      selectionMenuButtonsBuilder;
  final bool titleFullSpace;
  final Future<bool> Function() onWillPop;
  final bool itemsAreSelectable;
  final List<Widget> menuButtons;
  final Widget topWidget;
  final Widget noItemWidget;
  final bool showBackArrow;
  final Text selectionTitle;
  final IHelpWidget helpWidget;

  final bool isItemPicker;
  final bool multiSelectPicker;
  final List<DocumentReference> selectedItems;
  final void Function(List<DocumentReference> refs) saveRefsToItem;

  final ListScaffoldScrollBarConfiguration<T> scrollBar;

  const ListScaffoldWidget({
    Key key,
    this.title,
    this.itemBuilder,
    this.list,
    this.sortProperty,
    this.groupProperty,
    this.appBarButtons,
    this.selectionMenuButtonsBuilder,
    List<Widget> selectionMenuButtons,
    this.outsourceMenuButtons = false,
    this.groupTitle,
    this.titleFullSpace = false,
    this.itemsAreSelectable = true,
    this.menuButtons,
    this.topWidget,
    this.showBackArrow = false,
    this.reverseSort = false,
    this.sortProperties,
    this.noItemWidget,
    this.isItemPicker = false,
    this.multiSelectPicker,
    this.selectedItems,
    this.saveRefsToItem,
    this.selectionTitle,
    this.helpWidget,
    this.onWillPop,
    this.scrollBar,
  })  : assert(list != null),
        assert(itemBuilder != null),
        assert(sortProperty == null || sortProperties == null),
        assert(isItemPicker == false || multiSelectPicker != null),
        assert(isItemPicker == false || selectedItems != null),
        assert(isItemPicker == false || saveRefsToItem != null),
        super(key: key);

  @override
  ListScaffoldWidgetState<T> createState() => ListScaffoldWidgetState<T>();
}

class ListScaffoldWidgetState<T extends IdItemInterface>
    extends State<ListScaffoldWidget<T>> {
  bool selectionActive;
  List<T> selectedItems;

  List<List<T>> groupedItems;

  ScrollController scrollController = ScrollController();

  @override
  void initState() {
    super.initState();

    if (widget.isItemPicker) {
      selectedItems = widget.list
          .where((item) => widget.selectedItems.any((ref) => ref.id == item.id))
          .toList();
    } else
      selectedItems = <T>[];

    selectionActive = widget.isItemPicker;

    CleanAndroidDesignSettings.connectionChange?.listen((_) => setState(() {}));
  }

  @override
  Widget build(BuildContext context) {
    _sortList();

    var menuButtons = widget.isItemPicker
        ? _buildItemPickerMenuButtons()
        : (selectionActive && widget.selectionMenuButtonsBuilder != null
            ? widget.selectionMenuButtonsBuilder(context, this)
            : widget.menuButtons);
    var hasMenuButtons = menuButtons != null && menuButtons.length > 0;

    return DefaultScaffoldWidget(
      title: _getTitleWidget(),
      appBarButtons: selectionActive ? null : widget.appBarButtons,
      showBackArrow: widget.showBackArrow || selectionActive,
      outsourceMenuButtons: widget.outsourceMenuButtons,
      titleFullSpace: widget.titleFullSpace,
      onWillPop: () {
        if (selectionActive && !widget.isItemPicker) {
          cancelSelection();
          return Future.value(false);
        } else
          return widget.onWillPop?.call() ?? Future.value(true);
      },
      scrollable: false,
      helpWidget: widget.helpWidget,
      child: _buildChild(hasMenuButtons),
      menuButtons: menuButtons,
    );
  }

  Widget _buildChild(bool hasMenuButtons) {
    if (widget.list.length == 0) return _buildNoItemWidget();

    if (widget.scrollBar != null) {
      var sectionItems = <Widget>[];
      var sectionIndicatorItems =
          widget.scrollBar.groupScrollBarIndicatorWidget != null
              ? <Widget>[]
              : null;
      var sectionPositions = <double>[];
      var nextPos = widget.scrollBar.scrollBarOffset;
      groupedItems.forEach((group) {
        sectionItems.add(widget.scrollBar.groupScrollBarWidget(group.first));
        sectionIndicatorItems
            ?.add(widget.scrollBar.groupScrollBarIndicatorWidget(group.first));
        sectionPositions.add(nextPos);
        nextPos += widget.scrollBar.sectionHeaderHeight +
            group.length * widget.scrollBar.listItemHeight;
      });

      return Stack(
        children: [
          _getListView(),
          ScrollBarWidget(
            scrollBarConfiguration: widget.scrollBar,
            scrollController: scrollController,
            sectionItems: sectionItems,
            sectionIndicatorItems: sectionIndicatorItems,
            sectionPositions: sectionPositions,
            cutBottom: hasMenuButtons || widget.outsourceMenuButtons,
          ),
        ],
      );
    } else {
      return _getListView();
    }
  }

  List<Widget> _buildItemPickerMenuButtons() {
    return <Widget>[
      SimpleScaffoldMenuButtonWidget(
        text: Text("Abbrechen"),
        onTap: () => Navigator.pop(context),
        tooltip: "Auswahl abbrechen",
      ),
      SimpleScaffoldMenuButtonWidget(
        text: Text("Auswählen"),
        onTap: () => _savePickedItemsToItem(),
        tooltip: "Auswahl speichern",
      )
    ];
  }

  _savePickedItemsToItem() {
    var refList = selectedItems
        .map((item) => CleanAndroidDesignSettings.firestore
            .doc("${T.toString()}/${item.id}"))
        .toList();
    widget.saveRefsToItem(refList);

    Navigator.pop(context);
  }

  void _sortList() {
    if (widget.sortProperty != null)
      widget.list.sort((item1, item2) => widget
          .sortProperty(widget.reverseSort ? item2 : item1)
          .compareTo(widget.sortProperty(widget.reverseSort ? item1 : item2)));
    if (widget.sortProperties != null) {
      widget.list.sort((item1, item2) {
        var p1 = widget.sortProperties(widget.reverseSort ? item2 : item1);
        var p2 = widget.sortProperties(widget.reverseSort ? item1 : item2);
        for (var i = 0; i < p1.length; i++) {
          var result = p1[i].compareTo(p2[i]);
          if (result != 0) return result;
        }
        return 0;
      });
    }

    groupedItems = <List<T>>[];
    if (widget.groupProperty == null) {
      groupedItems.add(widget.list);
    } else if (widget.list.length > 0) {
      var lastList = <T>[];
      groupedItems.add(lastList);
      widget.list.asMap().forEach((i, item) {
        if (i == 0)
          lastList.add(item);
        else if (widget
                .groupProperty(lastList.last)
                .compareTo(widget.groupProperty(item)) ==
            0)
          lastList.add(item);
        else {
          lastList = <T>[];
          lastList.add(item);
          groupedItems.add(lastList);
        }
      });
    }
  }

  Widget _buildNoItemWidget() {
    return Column(
      children: <Widget>[
        widget.topWidget ?? Container(),
        DescriptionText.fromChild(widget.noItemWidget),
      ],
    );
  }

  Widget _getListView() {
    var offset = widget.topWidget == null ? 0 : 1;
    return ListView.builder(
      controller: scrollController,
      itemCount: groupedItems.length + offset,
      itemBuilder: (context, groupIndex) {
        if (widget.topWidget != null && groupIndex == 0)
          return widget.topWidget;

        return ItemSectionWidget.builder(
          title: widget.groupProperty == null
              ? null
              : widget.groupTitle == null
                  ? Text(widget
                      .groupProperty(groupedItems[groupIndex - offset].first)
                      .toString())
                  : widget.groupTitle(groupedItems[groupIndex - offset].first),
          itemCount: groupedItems[groupIndex - offset].length,
          itemBuilder: (context, itemIndex) {
            var item = groupedItems[groupIndex - offset][itemIndex];
            return widget.itemBuilder(context, this, item).copyWithSelection(
                isSelected: selectedItems.contains(item),
                onTapList: () {
                  if (selectionActive)
                    setState(() {
                      if (widget.isItemPicker && !widget.multiSelectPicker)
                        selectedItems.clear();
                      if (!selectedItems.remove(item)) selectedItems.add(item);
                    });
                  return !selectionActive;
                },
                onLongTapList: () {
                  if (widget.itemsAreSelectable)
                    setState(() {
                      selectionActive = true;
                      if (!selectedItems.remove(item)) selectedItems.add(item);
                    });
                });
          },
        );
      },
    );
  }

  Widget _getTitleWidget() {
    return selectionActive
        ? widget.selectionTitle ??
            Text(
              selectedItems.length.toString() + " ausgewählt",
            )
        : widget.title;
  }

  void cancelSelection() {
    setState(() {
      selectionActive = false;
      selectedItems.clear();
    });
  }
}
